//////////////////////////////////////////////////////////////////////////////
//
// AutoPitLot v 2.0
// Obsuga magnetometru MAG3110 na magistrali I2C
//
// (c) Pit Lab
// http://www.pitlab.pl
//////////////////////////////////////////////////////////////////////////////
#include "LPC1768.h"
#include "MAG3110.h"
#include "errors.h"
#include "sys_def.h"
#include "config_fram.h"
#include <math.h>


// Obsuga magnetometru wymaga wykonania kilku czynnoci rozoonych w czasie
// 1) Wystartowanie konwersji trwajcej 5ms (slot czasowy 0)
// 2) Wystartowanie odczytu wykonanego pomiaru. Odczyt wykonuje si w tle w procedurze przerwania I2C i 
//    trwa ok. 4ms (slot 11)
// 3) Przepisania odczytanych w przerwaniu danych do zmiennych pomiarowych

// Zmienne kalibracyjne
// 1) Zmienna FAH_MAG_MAG_CORR - korekcja zera mostka magnetometru kompensuje bdy nierwnomiernoci rozkadu 
//   pola w samym mostku. Uzyskuje si j przez odjecie od siebie wynikw pomiaru uzyskanych dla obu kierunkw
//   wstepnego namagnesownia struktury pomiarowej poleceniami SET i RESET. Zmiena = pomiar(SET) - pomiar(RESET)
// 2) Zmienna FAH_MAG_MAG_GAIN koryguje zaburzenie pola spododowane obecnoci ferromegnetykw obok czujnika
//   Powoduj one spaszczenie lub wyciagnicie pola w danej osi. Przemnoenie sygnau przez wspczynnik
//   daje jego normalizacj.
// 3) Zmiena FAH_MAG_MAG_OFST koryguje pooenie znormalizowanej wartoci pola magnetycznego, tak aby maksyma 
//   sygnau leay w rnych odlegociach od zera.  Zmienna = (Max + -Min)/2



////////////////////////////////////////////////////////////////////////////////
// Testuje obecno magnetometru MAG3110 na magistrali korzystajc z rejestru WHO_AM_I
// Parametry: 
// [i] chChan - kana magistrali I2C 0, 1 lub 2
// Zwraca: nic
// Czas zajcia magistrali I2C: ms przy zegarze 50kHz
////////////////////////////////////////////////////////////////////////////////
unsigned char TestPresenceMAG3110(unsigned char chChan)
{
    unsigned char chErr;
    volatile int nStartTime = T0TC;  //inicjuj pomiar czasu

    switch(chChan)
    {
    case I2C0: 
        chErr = StartI2C0(0x1C, WHO_AM_I, chData, 0, 1); 
        while (!chI2C0End)
            if (T0TC > nStartTime + SERVO_CNT_FREQ/(2*FREQ_SLOT_CZAS))   //czekaj maksymalnie 1/2 slotu czasowego
                return ERR_BRAK_MAGN;
        break;
    case I2C1:
        chErr = StartI2C1(0x1C, WHO_AM_I, chData, 0, 1);
        while (!chI2C1End)
            if (T0TC > nStartTime + SERVO_CNT_FREQ/(2*FREQ_SLOT_CZAS))   //czekaj maksymalnie 1/2 slotu czasowego
                return ERR_BRAK_MAGN;

        break;
    case I2C2:
        chErr = StartI2C2(0x1C, WHO_AM_I, chData, 0, 1);
        while (!chI2C2End)
            if (T0TC > nStartTime + SERVO_CNT_FREQ/(2*FREQ_SLOT_CZAS))   //czekaj maksymalnie 1/2 slotu czasowego
                return ERR_BRAK_MAGN;
        break;
    default: chErr = ERR_INVALID_DATA; break;
    }
    if (chData[0] != I_AM_MAG3110)
        chErr = ERR_BRAK_MAGN;

    return chErr;
}



////////////////////////////////////////////////////////////////////////////////
// Odczytuje temperatur struktury ukadu
// Parametry: 
// [i] chChan - kana magistrali I2C 0, 1 lub 2
// Zwraca: nic
// Czas zajcia magistrali I2C: 
////////////////////////////////////////////////////////////////////////////////
unsigned char ReadTemperatureMAG3110(unsigned char chChan, float* fTemp)
{
    unsigned char chErr;
    int nStartTime = T0TC;  //inicjuj pomiar czasu

    switch(chChan)
    {
    /*case I2C0:
        chErr = StartI2C0(0x1C, DIE_TEMP, chData, 0, 1); 
        while (!chI2C0End);
        break;
    case I2C1:
        chErr = StartI2C1(0x1C, DIE_TEMP, chData, 0, 1);
        while (!chI2C1End);
        break;*/
    case I2C2: 
        chErr = StartI2C2(0x1C, DIE_TEMP, chData, 0, 1);
        while (!chI2C2End)
            if (T0TC > nStartTime + SERVO_CNT_FREQ/(2*FREQ_SLOT_CZAS))   //czekaj maksymalnie 1/2 slotu czasowego
                return ERR_I2C_NOTREADY;
        break;
    default: chErr = ERR_INVALID_DATA; break;
    }
    if (chData[0] > 0x80)
        *fTemp = (float)(chData[0]-0x80)* (-1.0);
    else
        *fTemp = (float)chData[0];
     return chErr;
}


////////////////////////////////////////////////////////////////////////////////
// Budzi ukad i startuje konwersj magnetometru MAG3110
// Initiate a triggered measurement by writing 0b00011010 to CTRL_REG1 (CTRL_REG1 = 0b00011010).
// Dane powinny pojawi si 5ms po starcie konwersji
// Parametry: 
// [i] chChan - kana magistrali I2C 0, 1 lub 2
// Zwraca: nic
// Czas zajcia magistrali I2C: 0,80ms przy zegarze 38kHz
////////////////////////////////////////////////////////////////////////////////
unsigned char StartMagnetometerMAG(unsigned char chChan)
{
    unsigned char chErr;

    chData[0] = (0 << 0)|   //AC: Operating mode selection: 0: STANDBY mode; 1: ACTIVE mode.
                (1 << 1)|   //TM: Trigger immediate measurement: 1: praca kadorazowo wyzwalana
                (0 << 2)|   //FR: Fast Read selection. 0: The full 16-bit values are read.
                (2 << 3)|   //OS: over sampling ratio or measurement integration time: 4 sample
                (0 << 5);   //DR: Data rate selection: 000 najszybciej

    switch(chChan)
    {
    /*case I2C0: chErr = StartI2C0(0x1C, CTRL_REG1, chData, 1, 0); break;  //rozpocznij pomiar na I2C0        
    case I2C1: chErr = StartI2C1(0x1C, CTRL_REG1, chData, 1, 0); break;  //rozpocznij pomiar na I2C1*/
    case I2C2: chErr = StartI2C2(0x1C, CTRL_REG1, chData, 1, 0); break; //rozpocznij pomiar na I2C2
        
    default: chErr = ERR_INVALID_DATA; break;
    }
    return chErr;
}


////////////////////////////////////////////////////////////////////////////////
// Inicjuje odczyt danych z magnetometru MAG3110
// Parametry: 
// [i] chChan - kana magistrali I2C 1 lub 2
// Zwraca: nic
// Czas zajcia magistrali I2C: 2,28ms przy zegarze 38kHz
////////////////////////////////////////////////////////////////////////////////
unsigned char ReadMagnetometerMAG(unsigned char chChan)
{
    unsigned char chErr;

    switch(chChan)
    {
    /*case I2C0: chErr = StartI2C0(0x1C, OUT_X_MSB, chData, 0, 6); break;  //wylij polecenie zapisu wszystkich pomiarw
    case I2C1: chErr = StartI2C1(0x1C, OUT_X_MSB, chData, 0, 6); break;  //wylij polecenie zapisu wszystkich pomiarw        */
    case I2C2: chErr = StartI2C2(0x1C, OUT_X_MSB, chData, 0, 6); break;  //wylij polecenie odczytu wszystkich pomiarw
    default: chErr = ERR_INVALID_DATA; break;
    }
    return chErr;
}


////////////////////////////////////////////////////////////////////////////////
// Analizuje dane i liczy wartoci natenia pola magnetycznego we wszystkich 3 osiach
// Parametry: brak
// Zwraca: kod bdu
// Czas wykonania 27us
////////////////////////////////////////////////////////////////////////////////
unsigned char CalculateMagnetometerMAG(void)
{
    volatile short sTemp;
    unsigned char x;

    if ((!chData[0] && !chData[1]) | (!chData[2] && !chData[3]) | (!chData[4] && !chData[5]))
        return ERR_INVALID_DATA;

    //odejmij teoretyczny offset zera i filtruj dolnoprzepustowo
    sTemp = 0x100 * chData[0] + chData[1];
    fVectMagnMAG[0] = sTemp;
    sTemp = 0x100 * chData[2] + chData[3];
    fVectMagnMAG[1] = sTemp;   
    sTemp = 0x100 * chData[4] + chData[5];
    fVectMagnMAG[2] = -1 * sTemp;

    //jeeli trwa pomiar korekcji magnetometru na magistrali I2C1 to zwr surowe dane
    if (sInitFlag & FIND_CORR2)
        return ERR_OK;

    //jeeli trwa kalibracja to uwzgldniaj tylko wspczynnik korekcyjny
    if (sInitFlag & CALIB_MAG2)
        return ERR_OK;

    //normalny pomiar z uwzglednieniem wszystkich wsplczynikw korekcyjnych
    for (x=0; x<3; x++)
        fVectMagnMAG[x] = fVectMagnMAG[x] * fGainMagnMAG[x] - fOfstMagnMAG[x];     
    return ERR_OK;
}


////////////////////////////////////////////////////////////////////////////////
// Funkcja kapsukujca funkcj obliczeniow. Pozwala na uycie na zewntrz bez potrzeby publikowania danych
// Parametry: 
// [i] chChan - kana magistrali I2C 0, 1 lub 2
// Zwraca: kod bdu
////////////////////////////////////////////////////////////////////////////////
unsigned char MeasureMagnetometerMAG(unsigned char chChan)
{
    unsigned char x, chErr;

    switch(chChan)
    {
    case I2C2: 
        chErr = CalculateMagnetometerMAG();
        if (sInitFlag & CALIB_MAG2)     //jeeli trwa kalibracja to znajd wartoci maksymalne
        {
          for (x=0; x<3; x++)
          {
            if (fVectMagnMAG[x] > fMaxValMagn2[x])
            {
              fMaxValMagn2[x] = fVectMagnMAG[x];
              BeepMax();
            }
            if (fVectMagnMAG[x] < fMinValMagn2[x])
            {
              fMinValMagn2[x] = fVectMagnMAG[x];
              BeepMin();
            }
          }
        }
        break;

        default: 
            chErr = ERR_INVALID_DATA; 
            break;

    }
    return chErr;
}


////////////////////////////////////////////////////////////////////////////////
// Inicjuje zmienne konfiguracyjne
// Parametry: 
// Zwraca: kod bdu
////////////////////////////////////////////////////////////////////////////////
unsigned char InitMagnetometerMAG(void)
{
    unsigned char x, chErr = ERR_OK;   

    //sprawd obecno magnetometru
    chErr = TestPresenceMAG3110(2);
    if (chErr == ERR_BRAK_MAGN)
        return chErr;

    //Ustaw tryb pracy na 20Hz
    chData[0] = (1 << 0)|   //AC: Operating mode selection: 0: STANDBY mode; 1: ACTIVE mode.
                (1 << 1)|   //TM: Trigger immediate measurement: 1: praca kadorazowo wyzwalana
                (0 << 2)|   //FR: Fast Read selection. 0: The full 16-bit values are read.
                (2 << 3)|   //OS: over sampling ratio or measurement integration time: 4 sample
                (0 << 5);   //DR: Data rate selection: 000 najszybciej
    chErr = StartI2C2(0x1C, CTRL_REG1, chData, 1, 0);
    while (!chI2C2End);

    chData[0] = (0 << 4)|   //Magnetic Sensor Reset. 0: off,  1: on.
                (0 << 5)|   //Data output correction 0: off
                (1 << 7);   //Automatic Magnetic Sensor Reset. 0: off,  1: on.
    chErr = StartI2C2(0x1C, CTRL_REG2, chData, 1, 0);
    while (!chI2C2End);


    //czytaj temperatur struktury
    ReadTemperatureMAG3110(I2C2, &fTempMAG);


    //odczytaj wartoci kalibracyjne magnetometru
    for (x=0; x<3; x++)
    {
        //offset zera pocztkowego magnetometru
        chErr |= FramDataReadFloatValid((unsigned short)(FAH_MAG_MAG_CORRX+x*4), &fCorrMagnMAG[x], VALM_CORR_MAGN, VALP_CORR_MAGN, VALD_CORR_MAGN, ERR_MAGN_CAL_VAL);

        //wzmocnienie wskaza magnetometru
        chErr |= FramDataReadFloatValid((unsigned short)FAH_MAG_MAG_GAINX+x*4, &fGainMagnMAG[x], VALM_GAIN_MAGN, VALP_GAIN_MAGN, VALD_GAIN_MAGN, ERR_MAGN_CAL_VAL);
        
        //offset kompensacji hard/soft iron magnetometru
        chErr |= FramDataReadFloatValid((unsigned short)FAH_MAG_MAG_OFSTX+x*4, &fOfstMagnMAG[x], VALM_OFST_MAGN, VALP_OFST_MAGN, VALD_OFST_MAGN, ERR_MAGN_CAL_VAL);

    }

    if (!chErr)
        sInitFlag |= IDONE_MAG; //zakonczono kalibracj magnetometru
    return chErr;
}


////////////////////////////////////////////////////////////////////////////////
// Rozpoczyna i koczy kolejne etapy procedury kalibracyjnej magetometrw na obu magistralach
// Parametry: 
// [i] chEtap - etapy kalibracji: 
//  0 = zerowanie wspczynnikw korekcji na magistrali I2C1
//  1 = zerowanie wspczynnikw korekcji na magistrali I2C2
//  2 = ponowne wczytanie zewntrznie zapisanych zmiennych korekcji na magistrali I2C1
//  3 = ponowne wczytanie zewntrznie zapisanych zmiennych korekcji na magistrali I2C2
//  5 = ustawienie flagi kalibracji magnetometru na magistrali I2C1
//  6 = ustawienie flagi kalibracji magnetometru na magistrali I2C2
//  7 = kasowanie flagi kalibracji i obliczenie wspczynnikw kalibracji magnetometru na obu magistralach
// Zwraca: kod bdu
////////////////////////////////////////////////////////////////////////////////
unsigned char CalibrateMagnMAG(unsigned char chEtap)
{
    unsigned char x, chErr = ERR_OK;
    float fVect3D;
    
    switch (chEtap)
    {
   /* case 2:  //przeaduj konfiguracj  na kanale I2C1
        sInitFlag &= ~FIND_CORR1;
        for (x=0; x<3; x++)
        {
            chErr |= FramDataReadFloatValid(FAH_MAG_MAG_GAINX+x*4, &fGainMagnMAG[x], VALM_GAIN_MAGN, VALP_GAIN_MAGN, VALD_GAIN_MAGN, ERR_MAGN_CAL_VAL);           
            chErr |= FramDataReadFloatValid(FAH_MAG_MAG_OFSTX+x*4, &fOfstMagnMAG[x], VALM_OFST_MAGN, VALP_OFST_MAGN, VALD_OFST_MAGN, ERR_MAGN_CAL_VAL);            
        }
        break;*/

    case 3: //przeaduj konfiguracj  na kanale I2C2
        sInitFlag &= ~FIND_CORR2;
        for (x=0; x<3; x++)
        {
            chErr |= FramDataReadFloatValid(FAH_MAG_MAG_GAINX+x*4, &fGainMagnMAG[x], VALM_GAIN_MAGN, VALP_GAIN_MAGN, VALD_GAIN_MAGN, ERR_MAGN_CAL_VAL);
            chErr |= FramDataReadFloatValid(FAH_MAG_MAG_OFSTX+x*4, &fOfstMagnMAG[x], VALM_OFST_MAGN, VALP_OFST_MAGN, VALD_OFST_MAGN, ERR_MAGN_CAL_VAL);
        }
        break;

    /*case 5: //ustaw flag kalibracji magnetometru na magistrali I2C1
        sInitFlag |= CALIB_MAG1;
        for (x=0; x<3; x++)
        {
            fMinValMagn1[x] = fVectMagnMAG[x];
            fMaxValMagn1[x] = fVectMagnMAG[x];
        }
        break;*/

    case 6: //ustaw flag kalibracji magnetometru na magistrali I2C2
        sInitFlag |= CALIB_MAG2;
         for (x=0; x<3; x++)
        {
            fMinValMagn2[x] = fVectMagnMAG[x];
            fMaxValMagn2[x] = fVectMagnMAG[x];
        }
        break;

    case 7: //wykonaj kalibraj i kasuj flagi kalibracji magnetometru na magistralach I2C1 i I2C2
        /*if (sInitFlag & CALIB_MAG1)
        {
            //oblicz znormalizowan warto wektora 3D ekstremw pola magnetycznego
            fVect3D = sqrtf((fMaxValMagn1[0] - fMinValMagn1[0]) * (fMaxValMagn1[0] - fMinValMagn1[0])+
                            (fMaxValMagn1[1] - fMinValMagn1[1]) * (fMaxValMagn1[1] - fMinValMagn1[1])+
                            (fMaxValMagn1[2] - fMinValMagn1[2]) * (fMaxValMagn1[2] - fMinValMagn1[2]));

            //oblicz a nastpnie zapisz gain i ofset dla czujnika na magistrali I2C1
            for (x=0; x<3; x++)
            {
                fGainMagnMAG[x] = fVect3D / (fMaxValMagn1[x] - fMinValMagn1[x]);
                fOfstMagnMAG[x] = (fMaxValMagn1[x] + fMinValMagn1[x]) / 2 * fGainMagnMAG[x];
                FramDataWriteFloat(FAH_MAG_MAG_GAINX+x*4, fGainMagnMAG[x]);
                FramDataWriteFloat(FAH_MAG_MAG_OFSTX+x*4, fOfstMagnMAG[x]);
            }
            sInitFlag &= ~CALIB_MAG1;
        }*/
        if (sInitFlag & CALIB_MAG2)
        {
            //oblicz znormalizowan warto wektora 3D ekstremw pola magnetycznego
            fVect3D = sqrtf((fMaxValMagn2[0] - fMinValMagn2[0]) * (fMaxValMagn2[0] - fMinValMagn2[0])+
                            (fMaxValMagn2[1] - fMinValMagn2[1]) * (fMaxValMagn2[1] - fMinValMagn2[1])+
                            (fMaxValMagn2[2] - fMinValMagn2[2]) * (fMaxValMagn2[2] - fMinValMagn2[2]));

            //oblicz a nastpnie zapisz gain i ofset dla czujnika na magistrali I2C2
            for (x=0; x<3; x++)
            {
                fGainMagnMAG[x] = fVect3D / (fMaxValMagn2[x] - fMinValMagn2[x]);
                fOfstMagnMAG[x] = ((fMaxValMagn2[x] + fMinValMagn2[x]) / 2) * fGainMagnMAG[x];
                FramDataWriteFloat(FAH_MAG_MAG_GAINX+x*4, fGainMagnMAG[x]);
                FramDataWriteFloat(FAH_MAG_MAG_OFSTX+x*4, fOfstMagnMAG[x]);
            }
            sInitFlag &= ~CALIB_MAG2;
        }
        break;

    case 8: //zakocz kalibracj i korekcj na wszystkich kanaach
        sInitFlag &= ~(FIND_CORR1 | FIND_CORR2 | CALIB_MAG1 | CALIB_MAG2);
        break;
        
    case 9: //inicjalizuj dane do wartoci domylnych na I2C1
    case 10: //inicjalizuj dane do wartoci domylnych na I2C2
        for (x=0; x<3; x++)
        {
            fGainMagnMAG[x] = (float)1.0;
            fOfstMagnMAG[x] = (float)0.0;
            FramDataWriteFloat(FAH_MAG_MAG_GAINX+x*4, fGainMagnMAG[x]);
            FramDataWriteFloat(FAH_MAG_MAG_OFSTX+x*4, fOfstMagnMAG[x]);
        }
        break;

    default: 
        chErr = ERR_INVALID_DATA;
    }
    return chErr;
}
